"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Naming_1 = require("./Naming");
var Renderer_1 = require("./Renderer");
var ConvenienceRenderer = (function (_super) {
    __extends(ConvenienceRenderer, _super);
    function ConvenienceRenderer() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.nameForTopLevel = function (type, name) {
            var maybeNamedType = _this.namedTypeToNameForTopLevel(type);
            var styledName;
            if (maybeNamedType) {
                styledName = _this.namedTypeNameStyle(name);
            }
            else {
                styledName = _this.topLevelNameStyle(name);
            }
            var named = _this._globalNamespace.add(new Naming_1.FixedName(styledName));
            var dependencyNames = _this.topLevelDependencyNames(named);
            for (var _i = 0, dependencyNames_1 = dependencyNames; _i < dependencyNames_1.length; _i++) {
                var dn = dependencyNames_1[_i];
                _this._globalNamespace.add(dn);
            }
            if (maybeNamedType) {
                _this._classAndUnionNames = _this._classAndUnionNames.set(maybeNamedType, named);
            }
            return named;
        };
        _this.addClassOrUnionNamed = function (type) {
            if (_this._classAndUnionNames.has(type)) {
                return _this._classAndUnionNames.get(type);
            }
            var name = type.names.combined;
            var named = _this._globalNamespace.add(new Naming_1.SimpleName(_this.namedTypeNameStyle(name), _this.namedTypeNamer));
            _this._classAndUnionNames = _this._classAndUnionNames.set(type, named);
            return named;
        };
        _this.addPropertyNameds = function (c, classNamed) {
            var ns = new Naming_1.Namespace(c.names.combined, _this._globalNamespace, immutable_1.Set(), immutable_1.Set([classNamed]));
            var names = c.properties
                .map(function (t, name) {
                return ns.add(new Naming_1.SimpleName(_this.propertyNameStyle(name), _this.propertyNamer));
            })
                .toMap();
            _this._propertyNames = _this._propertyNames.set(c, names);
        };
        _this.childrenOfType = function (t) {
            var names = _this.names;
            if (t instanceof Type_1.ClassType) {
                var propertyNameds_1 = _this._propertyNames.get(t);
                return t.properties
                    .sortBy(function (_, n) { return names.get(propertyNameds_1.get(n)); })
                    .toOrderedSet();
            }
            return t.children.toOrderedSet();
        };
        _this.unionFieldName = function (t) {
            var typeNameForUnionMember = function (t) {
                if (t instanceof Type_1.PrimitiveType) {
                    switch (t.kind) {
                        case "any":
                            return "anything";
                        case "null":
                            return "null";
                        case "bool":
                            return "bool";
                        case "integer":
                            return "long";
                        case "double":
                            return "double";
                        case "string":
                            return "string";
                    }
                }
                else if (t instanceof Type_1.ArrayType) {
                    return typeNameForUnionMember(t.items) + "_array";
                }
                else if (t instanceof Type_1.ClassType) {
                    return _this.names.get(_this.nameForNamedType(t));
                }
                else if (t instanceof Type_1.MapType) {
                    return typeNameForUnionMember(t.values), "_map";
                }
                else if (t instanceof Type_1.UnionType) {
                    return "union";
                }
                throw "Unknown type";
            };
            return _this.propertyNameStyle(typeNameForUnionMember(t));
        };
        _this.nameForNamedType = function (t) {
            if (!_this._classAndUnionNames.has(t)) {
                throw "Named type does not exist.";
            }
            return _this._classAndUnionNames.get(t);
        };
        _this.forEachTopLevel = function (blankLocations, f) {
            _this.forEachWithBlankLines(_this.topLevels, blankLocations, function (t, name) {
                return f(t, _this._topLevelNames.get(name));
            });
        };
        _this.forEachClass = function (blankLocations, f) {
            // FIXME: sort property names
            _this.forEachWithBlankLines(_this._namedClasses, blankLocations, function (c) {
                return f(c, _this._classAndUnionNames.get(c), _this._propertyNames.get(c));
            });
        };
        _this.forEachUnion = function (blankLocations, f) {
            _this.forEachWithBlankLines(_this._namedUnions, blankLocations, function (u) {
                return f(u, _this._classAndUnionNames.get(u));
            });
        };
        return _this;
    }
    Object.defineProperty(ConvenienceRenderer.prototype, "forbiddenNames", {
        get: function () {
            return [];
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.topLevelDependencyNames = function (topLevelName) {
        return [];
    };
    ConvenienceRenderer.prototype.unionNeedsName = function (u) {
        return !Type_1.nullableFromUnion(u);
    };
    ConvenienceRenderer.prototype.setUpNaming = function () {
        var _this = this;
        this._globalNamespace = Naming_1.keywordNamespace("global", this.forbiddenNames);
        var _a = Type_1.allClassesAndUnions(this.topLevels), classes = _a.classes, unions = _a.unions;
        var namedUnions = unions.filter(function (u) { return _this.unionNeedsName(u); }).toOrderedSet();
        this._classAndUnionNames = immutable_1.Map();
        this._propertyNames = immutable_1.Map();
        this._topLevelNames = this.topLevels.map(this.nameForTopLevel).toMap();
        classes.forEach(function (c) {
            var named = _this.addClassOrUnionNamed(c);
            _this.addPropertyNameds(c, named);
        });
        namedUnions.forEach(function (u) { return _this.addClassOrUnionNamed(u); });
        return [this._globalNamespace];
    };
    Object.defineProperty(ConvenienceRenderer.prototype, "namedUnions", {
        get: function () {
            return this._namedUnions;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveUnions", {
        get: function () {
            return !this._namedUnions.isEmpty();
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.emitSource = function () {
        var _this = this;
        var _a = Type_1.allClassesAndUnions(this.topLevels, this.childrenOfType), classes = _a.classes, unions = _a.unions;
        this._namedClasses = classes;
        this._namedUnions = unions.filter(function (u) { return _this.unionNeedsName(u); }).toOrderedSet();
        this.emitSourceStructure();
    };
    return ConvenienceRenderer;
}(Renderer_1.Renderer));
exports.ConvenienceRenderer = ConvenienceRenderer;
