"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("lodash");
var immutable_1 = require("immutable");
var Type_1 = require("../Type");
var Support_1 = require("../Support");
var Naming_1 = require("../Naming");
var Renderer_1 = require("../Renderer");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var unicode = require("unicode-properties");
var SimpleTypesTargetLanguage = (function (_super) {
    __extends(SimpleTypesTargetLanguage, _super);
    function SimpleTypesTargetLanguage() {
        var _this = this;
        var declareUnionsOption = new RendererOptions_1.BooleanOption("declare-unions", "Declare unions as named types", false);
        _this = _super.call(this, "Simple Types", ["types"], "txt", [declareUnionsOption.definition]) || this;
        _this.declareUnionsOption = declareUnionsOption;
        return _this;
    }
    SimpleTypesTargetLanguage.prototype.renderGraph = function (topLevels, optionValues) {
        return new SimpleTypesRenderer(topLevels, !this.declareUnionsOption.getValue(optionValues)).render();
    };
    return SimpleTypesTargetLanguage;
}(TargetLanguage_1.TypeScriptTargetLanguage));
exports.default = SimpleTypesTargetLanguage;
function isStartCharacter(c) {
    return unicode.isAlphabetic(c.charCodeAt(0)) || c == "_";
}
function isPartCharacter(c) {
    var category = unicode.getCategory(c.charCodeAt(0));
    return _.includes(["Nd", "Pc", "Mn", "Mc"], category) || isStartCharacter(c);
}
var legalizeName = Support_1.legalizeCharacters(isPartCharacter);
function simpleNameStyle(original, uppercase) {
    return Support_1.startWithLetter(isStartCharacter, uppercase, Support_1.camelCase(legalizeName(original)));
}
var SimpleTypesRenderer = (function (_super) {
    __extends(SimpleTypesRenderer, _super);
    function SimpleTypesRenderer(topLevels, inlineUnions) {
        var _this = _super.call(this, topLevels) || this;
        _this.namedFromTopLevel = function (type, name) {
            // FIXME: leave the name as-is?
            var named = _this.namespace.add(new Naming_1.FixedName(simpleNameStyle(name, true)));
            var definedTypes = type.directlyReachableNamedTypes;
            if (definedTypes.size > 1) {
                throw "Cannot have more than one defined type per top-level";
            }
            // If the top-level type doesn't contain any classes or unions
            // we have to define a class just for the `FromJson` method, in
            // emitFromJsonForTopLevel.
            if (definedTypes.size === 1) {
                var definedType = definedTypes.first();
                _this.classAndUnionNames = _this.classAndUnionNames.set(definedType, named);
            }
            return named;
        };
        _this.addClassOrUnionNamed = function (type) {
            if (_this.classAndUnionNames.has(type)) {
                return _this.classAndUnionNames.get(type);
            }
            var name = type.names.combined;
            var named = _this.namespace.add(new Naming_1.SimpleName(simpleNameStyle(name, true)));
            _this.classAndUnionNames = _this.classAndUnionNames.set(type, named);
            return named;
        };
        _this.addPropertyNameds = function (c, classNamed) {
            var ns = new Naming_1.Namespace(c.names.combined, _this.namespace, immutable_1.Set(), immutable_1.Set([classNamed]));
            var nameds = c.properties
                .map(function (t, name) { return ns.add(new Naming_1.SimpleName(simpleNameStyle(name, false))); })
                .toMap();
            _this.propertyNames = _this.propertyNames.set(c, nameds);
        };
        _this.sourceFor = Type_1.matchTypeAll(function (anyType) { return "Any"; }, function (nullType) { return "Null"; }, function (boolType) { return "Bool"; }, function (integerType) { return "Int"; }, function (doubleType) { return "Double"; }, function (stringType) { return "String"; }, function (arrayType) { return ["List<", _this.sourceFor(arrayType.items), ">"]; }, function (classType) { return _this.classAndUnionNames.get(classType); }, function (mapType) { return ["Map<String, ", _this.sourceFor(mapType.values), ">"]; }, function (unionType) {
            var nullable = Type_1.nullableFromUnion(unionType);
            if (nullable)
                return ["Maybe<", _this.sourceFor(nullable), ">"];
            if (_this.inlineUnions) {
                var children = unionType.children.map(function (t) { return _this.sourceFor(t); });
                return Support_1.intercalate(" | ", children).toArray();
            }
            else {
                return _this.classAndUnionNames.get(unionType);
            }
        });
        _this.emitClass = function (c) {
            var propertyNames = _this.propertyNames.get(c);
            _this.emitLine("class ", _this.classAndUnionNames.get(c), " {");
            _this.indent(function () {
                _this.forEach(c.properties, false, false, function (t, name) {
                    _this.emitLine(propertyNames.get(name), ": ", _this.sourceFor(t));
                });
            });
            _this.emitLine("}");
        };
        _this.emitUnion = function (u) {
            _this.emitLine("union ", _this.classAndUnionNames.get(u), " {");
            _this.indent(function () {
                _this.forEach(u.children, false, false, function (t) {
                    _this.emitLine("case ", _this.sourceFor(t));
                });
            });
            _this.emitLine("}");
        };
        _this.childrenOfType = function (t) {
            if (t instanceof Type_1.ClassType) {
                var propertyNameds_1 = _this.propertyNames.get(t);
                return t.properties
                    .sortBy(function (_, n) { return _this.names.get(propertyNameds_1.get(n)); })
                    .toOrderedSet();
            }
            return t.children.toOrderedSet();
        };
        _this.inlineUnions = inlineUnions;
        return _this;
    }
    SimpleTypesRenderer.prototype.setUpNaming = function () {
        var _this = this;
        this.namespace = Naming_1.keywordNamespace("global", []);
        var _a = Type_1.allClassesAndUnions(this.topLevels), classes = _a.classes, unions = _a.unions;
        var namedUnions = unions.filter(function (u) { return !Type_1.nullableFromUnion(u); }).toSet();
        this.classAndUnionNames = immutable_1.Map();
        this.propertyNames = immutable_1.Map();
        this.topLevelNames = this.topLevels.map(this.namedFromTopLevel).toMap();
        classes.forEach(function (c) {
            var named = _this.addClassOrUnionNamed(c);
            _this.addPropertyNameds(c, named);
        });
        namedUnions.forEach(function (u) { return _this.addClassOrUnionNamed(u); });
        return [this.namespace];
    };
    SimpleTypesRenderer.prototype.emitSource = function () {
        var _a = Type_1.allClassesAndUnions(this.topLevels, this.childrenOfType), classes = _a.classes, unions = _a.unions;
        this.forEach(classes, true, false, this.emitClass);
        if (!this.inlineUnions) {
            var properUnions = unions.filter(function (u) { return !Type_1.nullableFromUnion(u); });
            this.forEach(properUnions, true, true, this.emitUnion);
        }
    };
    return SimpleTypesRenderer;
}(Renderer_1.Renderer));
