import { Set, OrderedSet, List, Map, Iterable } from "immutable";
export declare class Namespace {
    private readonly _name;
    private readonly _parent?;
    private _children;
    readonly forbiddenNamespaces: Set<Namespace>;
    readonly additionalForbidden: Set<Name>;
    private _members;
    constructor(name: string, parent: Namespace | undefined, forbiddenNamespaces: Set<Namespace>, additionalForbidden: Set<Name>);
    private addChild(child);
    readonly children: OrderedSet<Namespace>;
    readonly members: OrderedSet<Name>;
    readonly forbiddenNameds: Set<Name>;
    add<TName extends Name>(named: TName): TName;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare abstract class Namer {
    abstract name(proposedName: string, forbiddenNames: Set<string>, namesToAssign: Iterable<any, Name>): Map<Name, string>;
    abstract equals(other: any): boolean;
    abstract hashCode(): number;
}
export declare class PrefixNamer extends Namer {
    private readonly _prefixes;
    constructor(prefixes: string[]);
    name(proposedName: string, forbiddenNames: Set<string>, namesToAssign: Iterable<any, Name>): Map<Name, string>;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare function funPrefixNamer(upper: boolean): PrefixNamer;
export declare abstract class Name {
    readonly namingFunction: Namer | null;
    private _associates;
    constructor(namingFunction: Namer | null);
    equals(other: any): boolean;
    hashCode(): number;
    addAssociate(associate: AssociatedName): void;
    readonly abstract dependencies: List<Name>;
    isFixed(): this is FixedName;
    abstract proposeName(names: Map<Name, string>): string;
    nameAssignments(forbiddenNames: Set<string>, assignedName: string): Map<Name, string> | null;
}
export declare class FixedName extends Name {
    private readonly _fixedName;
    constructor(_fixedName: string);
    readonly dependencies: List<Name>;
    proposeName(names?: Map<Name, string>): string;
    hashCode(): number;
}
export declare class SimpleName extends Name {
    private readonly _proposed;
    private static defaultNamingFunction;
    constructor(_proposed: string, namingFunction?: Namer);
    readonly dependencies: List<Name>;
    proposeName(names?: Map<Name, string>): string;
    hashCode(): number;
}
export declare class AssociatedName extends Name {
    private readonly _sponsor;
    readonly getName: (sponsorName: string) => string;
    constructor(_sponsor: Name, getName: (sponsorName: string) => string);
    readonly dependencies: List<Name>;
    proposeName(names?: Map<Name, string>): string;
}
export declare class DependencyName extends Name {
    private readonly _dependencies;
    private readonly _proposeName;
    constructor(namingFunction: Namer, _dependencies: List<Name>, _proposeName: (names: List<string>) => string);
    readonly dependencies: List<Name>;
    proposeName(names: Map<Name, string>): string;
    hashCode(): number;
}
export declare function keywordNamespace(name: string, keywords: string[]): Namespace;
export declare function assignNames(rootNamespaces: OrderedSet<Namespace>): Map<Name, string>;
