"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var stringHash = require("string-hash");
var Support_1 = require("./Support");
var Namespace = (function () {
    function Namespace(name, parent, forbiddenNamespaces, additionalForbidden) {
        this._name = name;
        this.forbiddenNamespaces = forbiddenNamespaces;
        this.additionalForbidden = additionalForbidden;
        this._children = immutable_1.OrderedSet();
        this._members = immutable_1.OrderedSet();
        if (parent) {
            this._parent = parent;
            parent.addChild(this);
        }
    }
    Namespace.prototype.addChild = function (child) {
        this._children = this._children.add(child);
    };
    Object.defineProperty(Namespace.prototype, "children", {
        get: function () {
            return this._children;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Namespace.prototype, "members", {
        get: function () {
            return this._members;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Namespace.prototype, "forbiddenNameds", {
        get: function () {
            // FIXME: cache
            return (_a = this.additionalForbidden).union.apply(_a, this.forbiddenNamespaces.map(function (ns) { return ns.members.toSet(); }).toArray());
            var _a;
        },
        enumerable: true,
        configurable: true
    });
    Namespace.prototype.add = function (named) {
        this._members = this._members.add(named);
        return named;
    };
    Namespace.prototype.equals = function (other) {
        return this === other;
    };
    Namespace.prototype.hashCode = function () {
        var hash = stringHash(this._name);
        if (this._parent) {
            hash += this._parent.hashCode();
        }
        return hash | 0;
    };
    return Namespace;
}());
exports.Namespace = Namespace;
// `Namer`s are invoked to figure out what names to assign non-fixed `Name`s,
// and in particular to resolve conflicts.  Those arise under two circumstances,
// which can also combine:
//
// 1. A proposed name is the same as an already assigned name that's forbidden
//    for the name to be assigned.
// 2. There is more than one `Name` about to be assigned a name that all have
//    the same proposed name.
//
// The namer is invoked with the set of all assigned, forbidden names,
// the requested name, and the `Name`s to assign names to.
//
// `Namer` is a class so that we can compare namers and put them into immutable
// collections.
// FIXME: Put the target-specific name transformation function into the Namer,
// too.  Then a Name's `proposedName` can just be its `originalName` or whatever,
// and can do double-duty to specify user overriding when we implement it.
var Namer = (function () {
    function Namer() {
    }
    return Namer;
}());
exports.Namer = Namer;
var PrefixNamer = (function (_super) {
    __extends(PrefixNamer, _super);
    function PrefixNamer(prefixes) {
        var _this = _super.call(this) || this;
        _this._prefixes = immutable_1.OrderedSet(prefixes);
        return _this;
    }
    PrefixNamer.prototype.name = function (proposedName, forbiddenNames, namesToAssign) {
        if (namesToAssign.isEmpty()) {
            throw "Number of names can't be less than 1";
        }
        if (namesToAssign.size === 1) {
            var assignedForSingle = namesToAssign
                .first()
                .nameAssignments(forbiddenNames, proposedName);
            if (assignedForSingle) {
                return assignedForSingle;
            }
        }
        var allAssignedNames = immutable_1.Map();
        var prefixes = this._prefixes;
        var suffixNumber = 1;
        while (!namesToAssign.isEmpty()) {
            var name = namesToAssign.first();
            var nameToTry = void 0;
            if (!prefixes.isEmpty()) {
                nameToTry = prefixes.first() + proposedName;
                prefixes = prefixes.rest();
            }
            else {
                nameToTry = proposedName + suffixNumber.toString();
                suffixNumber++;
            }
            var assigned = name.nameAssignments(forbiddenNames, nameToTry);
            if (assigned === null)
                continue;
            allAssignedNames = allAssignedNames.merge(assigned);
            forbiddenNames = forbiddenNames.union(allAssignedNames);
            namesToAssign = namesToAssign.rest();
        }
        return allAssignedNames;
    };
    PrefixNamer.prototype.equals = function (other) {
        if (!(other instanceof PrefixNamer)) {
            return false;
        }
        return other._prefixes.equals(this._prefixes);
    };
    PrefixNamer.prototype.hashCode = function () {
        return this._prefixes.hashCode();
    };
    return PrefixNamer;
}(Namer));
exports.PrefixNamer = PrefixNamer;
var funPrefixes = [
    "Purple",
    "Fluffy",
    "Tentacled",
    "Sticky",
    "Indigo",
    "Indecent",
    "Hilarious",
    "Ambitious",
    "Cunning",
    "Magenta",
    "Frisky",
    "Mischievous",
    "Braggadocious"
];
function funPrefixNamer(upper) {
    var prefixes = upper ? funPrefixes : funPrefixes.map(Support_1.decapitalize);
    return new PrefixNamer(funPrefixes);
}
exports.funPrefixNamer = funPrefixNamer;
var Name = (function () {
    // If a Named is fixed, the namingFunction is null.
    function Name(namingFunction) {
        this.namingFunction = namingFunction;
        this._associates = immutable_1.Set();
    }
    Name.prototype.equals = function (other) {
        return this === other;
    };
    Name.prototype.hashCode = function () {
        return 0;
    };
    Name.prototype.addAssociate = function (associate) {
        this._associates = this._associates.add(associate);
    };
    Name.prototype.isFixed = function () {
        return this.namingFunction === null;
    };
    Name.prototype.nameAssignments = function (forbiddenNames, assignedName) {
        if (forbiddenNames.has(assignedName))
            return null;
        var assignments = immutable_1.Map().set(this, assignedName);
        var success = true;
        this._associates.forEach(function (an) {
            var associatedAssignedName = an.getName(assignedName);
            if (forbiddenNames.has(associatedAssignedName)) {
                success = false;
                return false;
            }
            assignments = assignments.set(an, associatedAssignedName);
        });
        if (!success)
            return null;
        return assignments;
    };
    return Name;
}());
exports.Name = Name;
// FIXME: FixedNameds should optionally be user-configurable
var FixedName = (function (_super) {
    __extends(FixedName, _super);
    function FixedName(_fixedName) {
        var _this = _super.call(this, null) || this;
        _this._fixedName = _fixedName;
        return _this;
    }
    Object.defineProperty(FixedName.prototype, "dependencies", {
        get: function () {
            return immutable_1.List();
        },
        enumerable: true,
        configurable: true
    });
    FixedName.prototype.proposeName = function (names) {
        return this._fixedName;
    };
    FixedName.prototype.hashCode = function () {
        return (_super.prototype.hashCode.call(this) + stringHash(this._fixedName)) | 0;
    };
    return FixedName;
}(Name));
exports.FixedName = FixedName;
var SimpleName = (function (_super) {
    __extends(SimpleName, _super);
    function SimpleName(_proposed, namingFunction) {
        var _this = _super.call(this, namingFunction || SimpleName.defaultNamingFunction) || this;
        _this._proposed = _proposed;
        return _this;
    }
    Object.defineProperty(SimpleName.prototype, "dependencies", {
        get: function () {
            return immutable_1.List();
        },
        enumerable: true,
        configurable: true
    });
    SimpleName.prototype.proposeName = function (names) {
        return this._proposed;
    };
    SimpleName.prototype.hashCode = function () {
        return (_super.prototype.hashCode.call(this) + stringHash(this._proposed)) | 0;
    };
    SimpleName.defaultNamingFunction = new PrefixNamer([]);
    return SimpleName;
}(Name));
exports.SimpleName = SimpleName;
var AssociatedName = (function (_super) {
    __extends(AssociatedName, _super);
    function AssociatedName(_sponsor, getName) {
        var _this = _super.call(this, null) || this;
        _this._sponsor = _sponsor;
        _this.getName = getName;
        return _this;
    }
    Object.defineProperty(AssociatedName.prototype, "dependencies", {
        get: function () {
            return immutable_1.List([this._sponsor]);
        },
        enumerable: true,
        configurable: true
    });
    AssociatedName.prototype.proposeName = function (names) {
        throw "AssociatedName must be assigned via its sponsor";
    };
    return AssociatedName;
}(Name));
exports.AssociatedName = AssociatedName;
var DependencyName = (function (_super) {
    __extends(DependencyName, _super);
    // _dependencies is a List as opposed to a set because it might contain
    // the same name more than once, and we don't want to put the burden of
    // checking on the renderer.
    // The `names` parameter of _proposeName will contain the names of all
    // `dependencies` in the same order as the latter.  If some of them are
    // the same, `names` will contain their names multiple times.
    function DependencyName(namingFunction, _dependencies, _proposeName) {
        var _this = _super.call(this, namingFunction) || this;
        _this._dependencies = _dependencies;
        _this._proposeName = _proposeName;
        return _this;
    }
    Object.defineProperty(DependencyName.prototype, "dependencies", {
        get: function () {
            return this._dependencies;
        },
        enumerable: true,
        configurable: true
    });
    DependencyName.prototype.proposeName = function (names) {
        var dependencyNames = this._dependencies.map(function (n) { return names.get(n); }).toList();
        return this._proposeName(dependencyNames);
    };
    DependencyName.prototype.hashCode = function () {
        return (_super.prototype.hashCode.call(this) + this._dependencies.hashCode()) | 0;
    };
    return DependencyName;
}(Name));
exports.DependencyName = DependencyName;
function keywordNamespace(name, keywords) {
    var ns = new Namespace(name, undefined, immutable_1.Set(), immutable_1.Set());
    for (var _i = 0, keywords_1 = keywords; _i < keywords_1.length; _i++) {
        var name_1 = keywords_1[_i];
        ns.add(new FixedName(name_1));
    }
    return ns;
}
exports.keywordNamespace = keywordNamespace;
function allNamespacesRecursively(namespaces) {
    return namespaces.union.apply(namespaces, namespaces.map(function (ns) { return allNamespacesRecursively(ns.children); }).toArray());
}
var NamingContext = (function () {
    function NamingContext(rootNamespaces) {
        var _this = this;
        this.names = immutable_1.Map();
        this._namedsForName = immutable_1.Map();
        this.isReadyToBeNamed = function (named) {
            if (_this.names.has(named))
                return false;
            return named.dependencies.every(function (n) { return _this.names.has(n); });
        };
        this.areForbiddensFullyNamed = function (namespace) {
            return namespace.forbiddenNameds.every(function (n) { return _this.names.has(n); });
        };
        this.isConflicting = function (named, namedNamespace, proposed) {
            // If the name is not assigned at all, there is no conflict.
            if (!_this._namedsForName.has(proposed))
                return false;
            // The name is assigned, but it might still not be forbidden.
            var conflicting;
            _this._namedsForName.get(proposed).forEach(function (n) {
                if (namedNamespace.members.contains(n) || namedNamespace.forbiddenNameds.contains(n)) {
                    conflicting = n;
                    return false;
                }
            });
            return !!conflicting;
        };
        this.assign = function (named, namedNamespace, name) {
            if (_this.names.has(named)) {
                throw "Named assigned twice";
            }
            if (_this.isConflicting(named, namedNamespace, name)) {
                throw "Assigned name conflicts";
            }
            _this.names = _this.names.set(named, name);
            if (!_this._namedsForName.has(name)) {
                _this._namedsForName = _this._namedsForName.set(name, immutable_1.Set());
            }
            _this._namedsForName.set(name, _this._namedsForName.get(name).add(named));
        };
        this.namespaces = allNamespacesRecursively(rootNamespaces);
    }
    return NamingContext;
}());
// Naming algorithm
function assignNames(rootNamespaces) {
    var ctx = new NamingContext(rootNamespaces);
    // Assign all fixed names.
    ctx.namespaces.forEach(function (ns) {
        return ns.members.forEach(function (n) {
            if (!n.isFixed())
                return;
            ctx.assign(n, ns, n.proposeName());
        });
    });
    var _loop_1 = function () {
        // 1. Find a namespace whose forbiddens are all fully named, and which has
        //    at least one unnamed Named that has all its dependencies satisfied.
        //    If no such namespace exists we're either done, or there's an unallowed
        //    cycle.
        var unfinishedNamespaces = ctx.namespaces.filter(ctx.areForbiddensFullyNamed);
        var readyNamespace = unfinishedNamespaces.find(function (ns) {
            return ns.members.some(ctx.isReadyToBeNamed);
        });
        if (!readyNamespace) {
            return { value: ctx.names };
        }
        var forbiddenNames = readyNamespace.members
            .toSet()
            .union(readyNamespace.forbiddenNameds)
            .filter(function (n) { return ctx.names.has(n); })
            .map(function (n) { return ctx.names.get(n); })
            .toSet();
        // 2. Sort those names into sets where all members of a set propose the same
        //    name and have the same naming function.
        var readyNames = readyNamespace.members.filter(ctx.isReadyToBeNamed);
        // It would be nice if we had tuples, then we wouldn't have to do this in
        // two steps.
        var byNamingFunction = readyNames.groupBy(function (n) { return n.namingFunction; });
        byNamingFunction.forEach(function (namedsForNamingFunction, namingFunction) {
            var byProposed = namedsForNamingFunction.groupBy(function (n) {
                return n.proposeName(ctx.names);
            });
            byProposed.forEach(function (nameds, proposed) {
                // 3. Use each set's naming function to name its members.
                var names = namingFunction.name(proposed, forbiddenNames, nameds);
                names.forEach(function (assigned, name) {
                    return ctx.assign(name, readyNamespace, assigned);
                });
                forbiddenNames = forbiddenNames.union(names.toSet());
            });
        });
    };
    for (;;) {
        var state_1 = _loop_1();
        if (typeof state_1 === "object")
            return state_1.value;
    }
}
exports.assignNames = assignNames;
