import { Map, Iterable } from "immutable";
import { TopLevels } from "./Type";
import { Name, Namespace } from "./Naming";
import { Source, Sourcelike } from "./Source";
import { Annotation } from "./Annotation";
export declare type RenderResult = {
    source: Source;
    names: Map<Name, string>;
};
export declare type BlankLineLocations = "none" | "leading-and-interposing";
export declare abstract class Renderer {
    protected readonly topLevels: TopLevels;
    private _names;
    private _lastNewline?;
    private _emitted;
    private _currentEmitTarget;
    constructor(topLevels: TopLevels);
    emitNewline(): void;
    emitLine(...lineParts: Sourcelike[]): void;
    emitAnnotated(annotation: Annotation, emitter: () => void): void;
    emitIssue(message: string, emitter: () => void): void;
    private changeIndent(offset);
    forEach<K, V>(iterable: Iterable<K, V>, interposedBlankLines: boolean, leadingBlankLine: boolean, emitter: (v: V, k: K) => void): void;
    forEachWithBlankLines<K, V>(iterable: Iterable<K, V>, blankLineLocations: BlankLineLocations, emitter: (v: V, k: K) => void): void;
    indent(fn: () => void): void;
    finishedSource: () => Source;
    protected abstract setUpNaming(): Namespace[];
    protected abstract emitSource(): void;
    render: () => RenderResult;
    readonly names: Map<Name, string>;
}
