"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Naming_1 = require("./Naming");
var Source_1 = require("./Source");
var Annotation_1 = require("./Annotation");
var Renderer = (function () {
    function Renderer(topLevels) {
        var _this = this;
        this.finishedSource = function () {
            return Source_1.sourcelikeToSource(_this._emitted);
        };
        this.render = function () {
            _this._names = Naming_1.assignNames(immutable_1.OrderedSet(_this.setUpNaming()));
            _this.emitSource();
            return { source: _this.finishedSource(), names: _this._names };
        };
        this.topLevels = topLevels;
        this._currentEmitTarget = this._emitted = [];
    }
    Renderer.prototype.emitNewline = function () {
        var nl = Source_1.newline();
        this._currentEmitTarget.push(nl);
        this._lastNewline = nl;
    };
    Renderer.prototype.emitLine = function () {
        var lineParts = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            lineParts[_i] = arguments[_i];
        }
        if (lineParts.length === 1) {
            this._currentEmitTarget.push(lineParts[0]);
        }
        else if (lineParts.length > 1) {
            this._currentEmitTarget.push(lineParts);
        }
        this.emitNewline();
    };
    Renderer.prototype.emitAnnotated = function (annotation, emitter) {
        var oldEmitTarget = this._currentEmitTarget;
        var emitTarget = [];
        this._currentEmitTarget = emitTarget;
        emitter();
        if (this._currentEmitTarget !== emitTarget) {
            throw "_currentEmitTarget not restored correctly";
        }
        this._currentEmitTarget = oldEmitTarget;
        var source = Source_1.sourcelikeToSource(emitTarget);
        this._currentEmitTarget.push(Source_1.annotated(annotation, source));
    };
    Renderer.prototype.emitIssue = function (message, emitter) {
        this.emitAnnotated(new Annotation_1.IssueAnnotation(message), emitter);
    };
    Renderer.prototype.changeIndent = function (offset) {
        if (!this._lastNewline) {
            throw "Cannot change indent for the first line";
        }
        this._lastNewline.indentationChange += offset;
    };
    Renderer.prototype.forEach = function (iterable, interposedBlankLines, leadingBlankLine, emitter) {
        var _this = this;
        var needBlank = false;
        iterable.forEach(function (v, k) {
            if (leadingBlankLine || (interposedBlankLines && needBlank)) {
                _this.emitNewline();
            }
            emitter(v, k);
            needBlank = true;
        });
    };
    Renderer.prototype.forEachWithBlankLines = function (iterable, blankLineLocations, emitter) {
        this.forEach(iterable, blankLineLocations != "none", blankLineLocations != "none", emitter);
    };
    Renderer.prototype.indent = function (fn) {
        this.changeIndent(1);
        fn();
        this.changeIndent(-1);
    };
    Object.defineProperty(Renderer.prototype, "names", {
        get: function () {
            if (this._names === undefined) {
                throw "Names accessed before they were assigned";
            }
            return this._names;
        },
        enumerable: true,
        configurable: true
    });
    return Renderer;
}());
exports.Renderer = Renderer;
