import { List } from "immutable";
import { Annotation } from "./Annotation";
import { Name } from "./Naming";
import { RenderResult } from "./Renderer";
export declare type Source = TextSource | NewlineSource | SequenceSource | AnnotatedSource | NameSource;
export interface TextSource {
    kind: "text";
    text: string;
}
export interface NewlineSource {
    kind: "newline";
    indentationChange: number;
}
export interface SequenceSource {
    kind: "sequence";
    sequence: List<Source>;
}
export interface AnnotatedSource {
    kind: "annotated";
    annotation: Annotation;
    source: Source;
}
export interface NameSource {
    kind: "name";
    named: Name;
}
export declare function newline(): NewlineSource;
export declare type Sourcelike = Source | string | Name | SourcelikeArray;
export interface SourcelikeArray extends Array<Sourcelike> {
}
export declare function sourcelikeToSource(sl: Sourcelike): Source;
export declare function annotated(annotation: Annotation, sl: Sourcelike): Source;
export interface Location {
    line: number;
    column: number;
}
export interface Span {
    start: Location;
    end: Location;
}
export interface SourceAnnotation {
    annotation: Annotation;
    span: Span;
}
export interface SerializedRenderResult {
    lines: string[];
    annotations: List<SourceAnnotation>;
}
export declare function serializeRenderResult({source, names}: RenderResult, indentation: string): SerializedRenderResult;
