"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Naming_1 = require("./Naming");
var Support_1 = require("./Support");
function newline() {
    // FIXME: use singleton?
    return { kind: "newline", indentationChange: 0 };
}
exports.newline = newline;
function sourcelikeToSource(sl) {
    if (sl instanceof Array) {
        return {
            kind: "sequence",
            sequence: immutable_1.List(sl.map(sourcelikeToSource))
        };
    }
    if (typeof sl === "string") {
        var lines = sl.split("\n");
        if (lines.length === 1) {
            return { kind: "text", text: sl };
        }
        return {
            kind: "sequence",
            sequence: Support_1.intercalate(newline(), immutable_1.List(lines).map(function (l) { return ({ kind: "text", text: l }); })).toList()
        };
    }
    if (sl instanceof Naming_1.Name) {
        return { kind: "name", named: sl };
    }
    return sl;
}
exports.sourcelikeToSource = sourcelikeToSource;
function annotated(annotation, sl) {
    return {
        kind: "annotated",
        annotation: annotation,
        source: sourcelikeToSource(sl)
    };
}
exports.annotated = annotated;
function assertNever(x) {
    throw new Error("Unexpected object: " + x);
}
function serializeRenderResult(_a, indentation) {
    var source = _a.source, names = _a.names;
    var indent = 0;
    var indentNeeded = 0;
    var lines = [];
    var currentLine = [];
    var annotations = [];
    function indentIfNeeded() {
        if (indentNeeded === 0)
            return;
        currentLine.push(indentation.repeat(indentNeeded));
        indentNeeded = 0;
    }
    function flattenCurrentLine() {
        var str = currentLine.join("");
        currentLine = [str];
        return str;
    }
    function currentLocation() {
        return { line: lines.length, column: flattenCurrentLine().length };
    }
    function finishLine() {
        lines.push(flattenCurrentLine());
        currentLine = [];
    }
    function serializeToStringArray(source, names) {
        switch (source.kind) {
            case "text":
                indentIfNeeded();
                currentLine.push(source.text);
                break;
            case "newline":
                finishLine();
                indent += source.indentationChange;
                indentNeeded = indent;
                break;
            case "sequence":
                source.sequence.forEach(function (s) { return serializeToStringArray(s, names); });
                break;
            case "annotated":
                var start = currentLocation();
                serializeToStringArray(source.source, names);
                var end = currentLocation();
                annotations.push({ annotation: source.annotation, span: { start: start, end: end } });
                break;
            case "name":
                if (!names.has(source.named)) {
                    throw "No name for Named";
                }
                indentIfNeeded();
                currentLine.push(names.get(source.named));
                break;
            default:
                return assertNever(source);
        }
    }
    serializeToStringArray(source, names);
    finishLine();
    return { lines: lines, annotations: immutable_1.List(annotations) };
}
exports.serializeRenderResult = serializeRenderResult;
